import os
from typing import Callable

from lenses import lens

from libtmux import formats, exc
from libtmux.window import Window as LTWindow

from amino import List, __, _, Maybe, Map
from amino.lazy import lazy

from ribosome.record import field, optional_field

from myo.ui.tmux.adapter import Adapter
from myo.ui.tmux.pane import PaneAdapter, Pane, NativePane
from myo.ui.tmux.layout import Layout
from myo.ui.tmux.util import parse_window_id
from myo.record import Named, Record
from myo.logging import Logging


class Size(Record):
    width = field(int)
    height = field(int)

    @staticmethod
    def create(width, height):
        return Size(width=width, height=height)

    @property
    def _str_extra(self):
        return List(self.width, self.height)

    @property
    def tuple(self):
        return self.width, self.height

    def __iter__(self):
        return iter((self.width, self.height))


class Window(Named):
    id = optional_field(int)
    root = field(Layout)
    size = optional_field(Size)

    @property
    def desc(self):
        id = self.id / ' @{}'.format | ''
        return 'W{}'.format(id)

    def pane_lens(self, f: Callable[[Pane], Maybe]):
        return self.root.pane_lens(f) / lens().root.add_lens

    @property
    def _str_extra(self):
        return super()._str_extra.cat(self.root)


class VimWindow(Window):
    window_name = '<vim>'

    def __new__(cls, *a, name=None, **kw):
        n = Maybe(name) | VimWindow.window_name
        return super().__new__(cls, *a, name=n, **kw)

    @property
    def desc(self):
        return 'V{}'.format(super().desc)


class NativeWindow(LTWindow, Logging):

    def __init__(self, session=None, **kwargs):
        if not session:
            raise ValueError('Window requires a Session, session=Session')
        self.session = session
        self.server = self.session.server
        if 'window_id' not in kwargs:
            raise ValueError('Window requires a `window_id`')
        self._window_id = kwargs['window_id']
        self._data = Map(kwargs) + ('id', self._window_id)

    @property
    def _info(self):
        return self._data

    def _list_panes(self):
        pformats = [
            'session_name', 'session_id',
            'window_index', 'window_id',
            'window_name'
        ] + formats.PANE_FORMATS
        tmux_formats = ['#{%s}\t' % f for f in pformats]
        proc = self.cmd('list-panes', '-F%s' % ''.join(tmux_formats))
        panes = proc.stdout
        pformats = [
            'session_name', 'session_id',
            'window_index', 'window_id', 'window_name'
        ] + formats.PANE_FORMATS
        panes = [dict(zip(
            pformats, window.split('\t'))) for window in panes]
        panes = [
            dict(
                (k, v) for k, v in window.items()
                if v or
                k == 'pane_current_path'
            )   # preserve pane_current_path, in case it entered a new process
                # where we may not get a cwd from.
            for window in panes
        ]
        return panes

    def list_panes(self):
        return [NativePane(window=self, **pane) for pane in self._panes]

    @property
    def attached_pane(self):
        return (List.wrap(self.panes)
                .find(lambda a: a._info.get('pane_active') == '1'))

    def split_window(
            self,
            target=None,
            start_directory=None,
            attach=True,
            vertical=True
    ):
        pformats = ['session_name', 'session_id',
                    'window_index', 'window_id'] + formats.PANE_FORMATS
        tmux_formats = ['#{%s}\t' % f for f in pformats]
        tmux_args = tuple()
        if target:
            tmux_args += ('-t%s' % target,)
        else:
            tmux_args += ('-t%s' % self.panes[0].get('pane_id'),)
        if vertical:
            tmux_args += ('-v',)
        else:
            tmux_args += ('-h',)
        tmux_args += (
            '-P', '-F%s' % ''.join(tmux_formats)
        )
        if start_directory:
            start_directory = os.path.expanduser(start_directory)
            tmux_args += ('-c%s' % start_directory,)
        if not attach:
            tmux_args += ('-d',)
        pane = self.cmd(
            'split-window',
            *tmux_args
        )
        if pane.stderr:
            raise exc.LibTmuxException(pane.stderr)
            if 'pane too small' in pane.stderr:
                pass
            raise exc.LibTmuxException(pane.stderr, self._info, self.panes)
        else:
            pane = pane.stdout[0]
            pane = dict(zip(pformats, pane.split('\t')))
            pane = dict((k, v) for k, v in pane.items() if v)
        return NativePane(window=self, **pane)


class WindowAdapter(Adapter):

    @lazy
    def panes(self):
        return List.wrap(self.native.panes) / PaneAdapter

    def pane_by_id(self, id: int):
        return self.panes.find(__.id_i.contains(id))

    @lazy
    def id(self):
        return self.native._window_id

    @lazy
    def id_i(self):
        return parse_window_id(self.id)

    @lazy
    def size(self):
        return Size.create(int(self.native.width), int(self.native.height))

    @property
    def active_pane(self):
        return self.panes.find(_.active)

    def kill(self):
        self.native.kill_window()

    @property
    def name(self):
        return self.native.name

__all__ = ('WindowAdapter', 'Window')



from subprocess import Popen, PIPE, STDOUT
from typing import Union

from libtmux.exc import LibTmuxException

from amino.test.path import fixture_path
from amino import _, __, List, Maybe
from ribosome.test.integration.klk import later

from kallikrein.matchers import contain
from kallikrein import kf, Expectation, k
from amino.lazy import lazy
from amino.test.spec import SpecBase

from myo.ui.tmux.server import Server, NativeServer
from myo.ui.tmux.pane import PaneAdapter
from myo.ui.tmux.window import WindowAdapter
from myo.ui.tmux.session import SessionAdapter


class Spec(SpecBase):
    pass


class TmuxSpecBase(Spec):

    @property
    def win_width(self) -> int:
        return 100

    @property
    def win_height(self) -> int:
        return 40

    @property
    def socket(self) -> str:
        return 'myo_spec'

    @property
    def _socket_name(self) -> Union[str, None]:
        return self.socket if self.external_server else None

    @property
    def native_server(self) -> NativeServer:
        return NativeServer(socket_name=self._socket_name)

    def _find_server(self) -> bool:
        try:
            return self.native_server is not None
        except LibTmuxException:
            return False

    def _find_session(self) -> bool:
        try:
            return self.session is not None
        except LibTmuxException:
            return False

    @property
    def external_server(self) -> bool:
        return True

    @property
    def external_terminal(self) -> bool:
        return True

    @property
    def _terminal_args(self) -> list:
        geom = '{}x{}'.format(self.win_width + 1, self.win_height + 1)
        return ['urxvt', '-geometry', geom, '-e']

    @property
    def _tmux_args(self) -> list:
        conf = fixture_path('conf', 'tmux.conf')
        return ['tmux', '-L', self.socket, '-f', str(conf)]

    @lazy
    def term(self) -> Popen:
        t = self._terminal_args if self.external_terminal else []
        args = t + self._tmux_args
        return Popen(args, stdout=PIPE, stderr=STDOUT)

    def _setup_server(self) -> None:
        if self.external_server:
            self.term
            self._wait_for(self._find_server)
        self._wait_for(self._find_session)

    def _teardown_server(self) -> None:
        if self.external_server:
            self.term.kill()
            self.server.kill()

    @property
    def server(self) -> Server:
        return Server(self.native_server)

    @property
    def sessions(self) -> List[SessionAdapter]:
        return self.server.sessions

    @property
    def session(self) -> SessionAdapter:
        if self.external_server:
            return self.sessions[0]
        else:
            return (self.sessions.find(_.attached)
                    .get_or_fail('no attached session'))

    @property
    def _window(self) -> WindowAdapter:
        return (self.session.windows.last).get_or_fail('no window')

    @property
    def _panes(self) -> List[PaneAdapter]:
        return self._window.panes

    def _pane_with_id(self, id: int) -> Maybe[PaneAdapter]:
        return self._panes.find(__.id_i.contains(id))

    @property
    def _sizes(self) -> List[int]:
        return self._panes / _.size

    @property
    def _output(self) -> List[str]:
        return self._panes // _.capture

    def _output_contains(self, target: str) -> Expectation:
        return later(kf(lambda: self._output).must(contain(target)))

    def _pane_output(self, id: int) -> Maybe[List[str]]:
        return self._pane_with_id(id) / _.capture

    def _pane_output_contains(self, id: int, data: str) -> Expectation:
        def output() -> List[str]:
            return self._pane_output(id) | List()
        return later(kf(output).must(contain(data)))

    def _pane_output_contains_not(self, id: int, data: str) -> Expectation:
        self._wait(1)
        output = self._pane_output(id)
        return k(output).must(contain(~contain(data)))

__all__ = ('Spec', 'TmuxSpecBase')



from amino import IO, L, Just, _

from ribosome.unite import UniteSource


class UniteNames():
    history_candidates = '_myo_unite_history'
    commands_candidates = '_myo_unite_commands'
    run = '_myo_unite_run_command'
    delete = '_myo_unite_delete_command'
    commands = 'myo_commands'
    command = 'myo_command'
    history = 'myo_history'
    history_command = 'myo_history_command'
    syntax = '_myo_unite_syntax'


def commands_syntax(syntax):
    m = L(IO.call)(syntax.match, _, _)
    hi = L(IO.call)(syntax.highlight, _)
    name = '^\s*\[\S\+\]'
    return (
        m('name', name) +
        hi('name', ctermfg=3) +
        m('line', '\({} \)\@<=.*'.format(name)) +
        hi('line', ctermfg=4)
    )


class HistorySource(UniteSource):

    def __init__(self) -> None:
        super().__init__(UniteNames.history, UniteNames.history_candidates,
                         UniteNames.history_command, Just(UniteNames.syntax))

    def syntax_task(self, syntax):
        return commands_syntax(syntax)


class CommandsSource(UniteSource):

    def __init__(self) -> None:
        super().__init__(UniteNames.commands, UniteNames.commands_candidates,
                         UniteNames.command, Just(UniteNames.syntax))

    def syntax_task(self, syntax):
        return commands_syntax(syntax)

__all__ = ('HistorySource', 'UniteNames', 'CommandsSource')



from functools import singledispatch

from ribosome.machine.messages import Error

from myo.dispatch import Dispatcher
from myo.command import ShellCommand, CommandJob
from myo.components.tmux.message import TmuxRunCommand, TmuxRebootCommand
from myo.components.command.main import Reboot


@singledispatch
def convert_message(msg, options):
    return Error('invalid message for tmux dispatch: {}'.format(msg))


@convert_message.register(ShellCommand)
def _convert_shell_command(cmd, options):
    return TmuxRunCommand(CommandJob(command=cmd, options=options), options)


@convert_message.register(CommandJob)
def _convert_command_job(cmd, options):
    return TmuxRunCommand(cmd, options)


@convert_message.register(Reboot)
def _convert_command_reboot(cmd, options):
    return TmuxRebootCommand(cmd.job, options)


class TmuxDispatcher(Dispatcher):

    def can_run(self, cmd):
        return isinstance(cmd, (ShellCommand, CommandJob, Reboot))

    def message(self, cmd, options):
        return convert_message(cmd, options)

__all__ = ('TmuxDispatcher',)



import asyncio

from psutil import Process

from ribosome.record import list_field, bool_field, field
from ribosome.machine.transition import may_handle, handle, Transitions
from ribosome.machine.message_base import message
from ribosome.machine.state import StateMachine
from ribosome.machine.base import UnitIO

from amino import Empty, Try, __, List, L, _
from amino.io import IO

from myo.record import Record
from myo.components.tmux.message import WatchCommand
from myo.command import CommandJob
from myo.ui.tmux.pane import Pane


class WatcherState(Record):
    commands = list_field()
    watching = bool_field()

    def remove(self, cmds):
        return self.set(commands=self.commands.remove_all(cmds))


class WatchedCommand(Record):
    command = field(CommandJob)
    pane = field(Pane)


Start = message('Start')
Stop = message('Stop')
Tick = message('Tick')
Tail = message('Tail', 'pane')
Terminated = message('Terminated', 'command', 'pane')


class WatcherTransitions(Transitions):

    @property
    async def _sleep(self):
        await asyncio.sleep(self.machine.interval)
        return Empty()

    @may_handle(WatchCommand)
    def watch_command(self):
        pane = self.msg.pane
        wc = WatchedCommand(command=self.msg.command, pane=pane)
        return self.data.append1.commands(wc), Tail(pane), Start()

    @handle(Start)
    def start_rec(self):
        return self.data.watching.no.maybe_call(
            lambda: (self.data.set(watching=True), Tick().pub))

    @may_handle(Stop)
    def stop_rec(self):
        return self.data.set(watching=False)

    @handle(Tick)
    def tick(self):
        return self.data.watching.maybe_call(self._tick)

    @may_handle(Tail)
    def tail(self):
        return

    def _tick(self):
        restart = List(self._sleep, Tick().pub)
        done = self.data.commands.flat_map(self._check_proc)
        data = self.data.remove(done)
        term_msg = lambda wc: Terminated(wc.command, wc.pane)
        term = (done / term_msg / L(IO.call)(self.machine.bubble, _) /
                UnitIO)
        return term.cons(data) + restart

    def _check_proc(self, wc):
        proc = wc.pane.pid // (lambda a: Try(Process, a).to_maybe)
        return proc.exists(__.is_running()).no.maybe(wc)


class Watcher(StateMachine):
    _data_type = WatcherState

    Transitions = WatcherTransitions

    def __init__(self, machine, interval: float) -> None:
        super().__init__(List(), parent=machine)
        self.machine = machine
        self.interval = interval

    @property
    def title(self):
        return 'tmux watcher'

    @property
    def init(self):
        return WatcherState()

__all__ = ('Watcher',)



import amino.logging
from ribosome.logging import ribosome_logger

from amino.lazy import lazy

log = myo_root_logger = ribosome_logger('myo')


def myo_logger(name: str) -> amino.logging.Logger:
    return myo_root_logger.getChild(name)


class Logging(amino.logging.Logging):

    @lazy
    def _log(self) -> amino.logging.Logger:
        return myo_logger(self.__class__.__name__)

__all__ = ('myo_logger', 'Logging')



from ribosome.settings import PluginSettings, state_dir_help_default

state_dir_help = f'''{state_dir_help_default}
Stored data consists of:
* command history
* last test command line
* last shell used for testing
* layout state
'''


class ProteomeSettings(PluginSettings):

    def __init__(self) -> None:
        super().__init__('proteome', state_dir_help)

__all__ = ('ProteomeSettings',)



import abc
from typing import Callable

from amino import List, Path, Either, __, L, _, Just, Maybe, Map, Empty
from amino.io import IO
from amino.lazy import lazy

from ribosome import NvimFacade
from ribosome.machine.state import RunScratchMachine, IfUnhandled
from ribosome.machine.messages import Nop
from ribosome.machine.message_base import Message
from ribosome.util.callback import VimCallback

from myo.logging import Logging
from myo.output.data import ParseResult
from myo.output.machine import OutputMachine, SetResult


class OutputHandler(Logging, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def parse(self, output: List[str], errfile: Path) -> IO[ParseResult]:
        ...

    @abc.abstractmethod
    def display(self, result: ParseResult, jump: Maybe[Callable]
                ) -> IO[List[Message]]:
        ...


# TODO in `display`, return message like `DispalayCustomOutput` and move the
# `IfUnhandled` part to the core machine, using transition priorities
class CustomOutputHandler(OutputHandler):

    def __init__(self, vim: NvimFacade, handler: Callable[[str], ParseResult]
                 ) -> None:
        self.vim = vim
        self.handler = handler

    def parse(self, output: List[str], errfile: Path):
        return IO.call(self.handler, output)

    def display(self, result: ParseResult, options: Map[str, str]):
        ctor = L(OutputMachine)(self.vim, _, result, _, options)
        size = self.vim.vars.p('scratch_size') | 10
        msg = SetResult(result, options)
        opt = Map(use_tab=False, size=Just(size), wrap=True, init=msg)
        run = RunScratchMachine(ctor, options=opt)
        return IO.just(IfUnhandled(msg, run).pub)


class VimCompiler(OutputHandler, VimCallback):

    def __call__(self, data):
        return Empty()

    def parse(self, output: List[str], errfile: Path):
        r = ParseResult(head=List('errorformat'))
        return (IO.call(self.vim.cmd_sync, 'cgetfile {}'.format(errfile))
                .replace(r))

    def display(self, result, jump):
        copen = IO.call(self.vim.cmd_sync, 'copen')
        cfirst = IO.call(self.vim.cmd_sync, 'cfirst')
        return (copen + cfirst).replace(Just(Nop()))


class Parsing(CustomOutputHandler):

    def __init__(self, vim: NvimFacade, langs: List[str]) -> None:
        self.vim = vim
        self.langs = langs

    @lazy
    def parsers(self):
        mod = 'myo.output.parser'
        def create(lang):
            return Either.import_name('{}.{}'.format(mod, lang),
                                      'Parser').to_list
        return self.langs // create / (lambda a: a())

    def parse(self, output: List[str], errfile: Path):
        events = self.parsers // __.events(output)
        return IO.now(ParseResult(events=events, langs=self.langs))

__all__ = ('CustomOutputHandler', 'VimCompiler', 'Parsing')



from uuid import UUID

from ribosome.record import Record as RecordBase, uuid_field, field

from amino import List

from myo.util import Ident
from myo.util.ident import Key


class Record(RecordBase):
    uuid = uuid_field()

    @property
    def ident(self) -> Ident:
        return self.uuid


class Named(Record):
    name = field(str)

    def has_uuid(self, uuid: UUID):
        return self.uuid == uuid

    def has_name(self, name: str):
        return self.name == name

    def has_key(self, key: Key):
        return self.has_uuid(key.uuid) or self.has_name(key.name)

    def has_ident(self, ident: Ident):
        matcher = (
            self.has_uuid
            if isinstance(ident, UUID) else
            self.has_key
            if isinstance(ident, Key) else
            self.has_name
        )
        return matcher(ident)

    @property
    def _str_extra(self):
        return List(self.name)

__all__ = ('Record',)



from functools import singledispatch  # type: ignore
from operator import ne, sub

from amino import Either, IO, L, _, Right, __, List, Boolean, Empty

from myo.logging import Logging
from myo.ui.tmux.window import Window
from myo.ui.tmux.pane import Pane
from myo.ui.tmux.layout import Layout
from myo.ui.tmux.facade.window import WindowFacade
from myo.ui.tmux.view import View


class WindowPacker(Logging):

    def __init__(self, window: WindowFacade) -> None:
        self.window = window

    @property
    def run(self) -> IO[Either[str, Window]]:
        w, h = self.window.window_size
        return (
            self._pack_layout(self.window.root, w, h) /
            Right /
            __.replace(self.window)
        )

    def _loc(self, pane: Pane):
        return self.window.loc(pane)

    def _pack_layout(self, l, w, h):
        nop = IO.now(Empty())
        horizontal = l.horizontal
        total = w if horizontal else h
        @singledispatch
        def recurse(v, size, multi):
            pass
        @recurse.register(Pane)
        def rec_pane(v, size, multi):
            return self._apply_size(v, size, horizontal) if multi else nop
        @recurse.register(Layout)
        def rec_layout(v, size, multi):
            sub_size = (size, h) if horizontal else (w, size)
            return (
                self._pack_layout(v, *sub_size)
                .flat_replace(
                    self.window.ref_pane_fatal(v) //
                    L(self._apply_size)(_, size, horizontal) if multi else nop
                )
            )
        views = self.window.open_views(l)
        count = views.length
        if count > 0:
            # each pane spacer takes up one cell
            m = self._measure_layout(views, total - count + 1)
            return (
                self._apply_positions(views, horizontal) +
                views.zip(m).map2(L(recurse)(_, _, count > 1)).sequence(IO)
            )
        else:
            return nop

    def _measure_layout(self, views, total):
        calc = lambda s: s if s > 1 else s * total
        min_s = _actual_min_sizes(views) / calc
        max_s = _actual_max_sizes(views) / calc
        weights = _weights(views)
        return self._balance_sizes(min_s, max_s, weights, total) / round

    def _balance_sizes(self, min_s, max_s, weights, total):
        return _rectify_sizes(
            self._cut_sizes(min_s, weights, total) if sum(min_s) > total
            else self._distribute_sizes(min_s, max_s, weights, total)
        )

    def _cut_sizes(self, min_s, weights, total):
        self.log.debug('cut sizes: min {}, weights {}, total {}'.format(
            min_s, weights, total))
        surplus = sum(min_s) - total
        dist = _reverse_weights(weights) / (surplus * _)
        cut = (min_s.zip(dist)).map2(sub)
        neg = (cut / (lambda a: a if a < 0 else 0))
        neg_total = sum(neg)
        neg_count = neg.filter(_ < 0).length
        dist2 = neg_total / (min_s.length - neg_count)
        return cut / (lambda a: 0 if a < 0 else a + dist2)

    def _distribute_sizes(self, min_s, max_s, weights, total):
        self.log.debug(
            'distribute sizes: min {}, max {}, weights {}, total {}'.format(
                min_s, max_s, weights, total
            ))
        count = max_s.length
        surplus = total - sum(min_s)
        sizes1 = min_s.zip(max_s, weights).map3(
            lambda l, h, w: min(l + w * surplus, h))
        rest = total - sum(sizes1)
        unsat = (max_s.zip(sizes1)).map2(ne) / Boolean
        unsat_left = count - sum(unsat / _.value) > 0
        def trim_weights():
            w1 = (unsat.zip(weights)).map2(lambda s, w: s.maybe(w))
            return _normalize_weights(w1)
        def dist_rest():
            rest_w = trim_weights() if unsat_left else weights
            return (sizes1.zip(rest_w)).map2(lambda a, w: a + w * rest)
        return sizes1 if rest <= 0 else dist_rest()

    def _apply_size(self, pane, size, horizontal):
        self.log.debug('resize {} to {} ({})'.format(pane, size,
                                                     bool(horizontal)))
        return (
            self.window.native_pane_task_fatal(pane) //
            __.resize(size, horizontal)
        )

    def _apply_positions(self, views, horizontal):
        if views.length > 1 and views.exists(_.position.present):
            quantity = _.left if horizontal else _.top
            ordered = views.sort_by(_.position | 0).reversed
            adapters = self._ref_adapters(ordered)
            current = adapters.sort_by(lambda a: quantity(a) | 0)
            def setpos(current, correct):
                def rec(head, tail):
                    def swap(last, init):
                        if last != head:
                            last.swap(head)
                            return setpos(init.replace_item(head, last), tail)
                        else:
                            return setpos(init, tail)
                    return current.detach_last.map2(swap)
                return correct.detach_head.map2(rec)
            return IO.call(setpos, current, adapters)
        else:
            return IO.now(Empty())

    def _ref_adapters(self, views):
        return views // self._ref_adapter

    def _ref_adapter(self, view: View):
        return (self.window.native_pane(view) if isinstance(view, Pane) else
                self.window.native_ref_pane(view))


def _pos(a):
    return max(a, 0)


def _reverse_weights(weights):
    r = weights / (1 - _)
    norm = sum(r)
    return r / (_ / norm) if norm > 0 else r


def _rectify_sizes(sizes: List[int]):
    pos = sizes / _pos
    pos_count = sizes.filter(_ >= 2).length
    unders = pos / (lambda a: 2 - a if a < 2 else 0)
    sub = sum(unders) / pos_count
    return pos / (lambda a: 2 if a < 2 else max(2, a - sub))


def _actual_min_sizes(views):
    return views / (lambda a: a.effective_fixed_size.or_else(a.min_size) |
                    0)


def _actual_max_sizes(views):
    return views / (lambda a: a.effective_fixed_size.or_else(a.max_size) |
                    999)


def _weights(views):
    return _normalize_weights(views / _.effective_weight)


def _normalize_weights(weights):
    amended = amend_weights(weights)
    total = sum(amended)
    total1 = 1 if total == 0 else total
    return amended / (_ / total1)


def amend_weights(weights):
    total = sum(weights.join)
    total1 = 1 if total == 0 else total
    empties = weights.filter(_.is_empty).length
    empties1 = 1 if empties == 0 else empties
    empty = total1 / empties1
    return weights / (_ | empty)

__all__ = ('WindowPacker',)



from typing import Callable

from amino import List, _, __, Just, Map, Right, Empty, Either, Boolean, I, Maybe, L, Nil
from amino.lazy import lazy
from amino.io import IO

from ribosome.machine.base import UnitIO, DataIO
from ribosome.machine.transition import may_handle, handle, Error, Fatal
from ribosome.machine.messages import Stage1, Quit, Nop, Message
from ribosome.machine.state import Component
from ribosome.nvim.components import Flags

from myo.components.tmux.message import (TmuxOpen, TmuxRunCommand, TmuxCreatePane,
                                      TmuxCreateLayout, TmuxCreateSession,
                                      TmuxSpawnSession, TmuxFindVim, TmuxInfo,
                                      TmuxLoadDefaults, TmuxClosePane,
                                      TmuxRunLineInShell, StartWatcher,
                                      WatchCommand, QuitWatcher, TmuxPack,
                                      TmuxPostOpen, TmuxFixFocus, TmuxMinimize,
                                      TmuxRestore, TmuxToggle, TmuxFocus,
                                      TmuxOpenOrToggle, TmuxKill,
                                      UpdateVimWindow, TmuxRunTransient,
                                      TmuxPostCommand, TmuxRebootCommand)
from myo.ui.tmux.pane import Pane, VimPane, PaneData
from myo.ui.tmux.layout import LayoutDirections, Layout, VimLayout
from myo.ui.tmux.session import Session, VimSession
from myo.ui.tmux.server import Server, NativeServer
from myo.util import Ident
from myo.ui.tmux.window import VimWindow
from myo.components.core.message import AddDispatcher, Resized, Initialized
from myo.components.tmux.dispatch import TmuxDispatcher
from myo.ui.tmux.view_path import LayoutPathMod, PanePathMod, ViewPathMod
from myo.ui.tmux.data import TmuxState
from myo.components.command.message import SetShellTarget, CommandExecuted
from myo.command import Command, Shell, default_signals, CommandJob
from myo.components.tmux.watcher import Watcher, Terminated
from myo.ui.tmux.facade.main import TmuxFacade
from myo.ui.tmux.util import format_state


invalid_pane_name = 'invalid pane name: {}'


class Tmux(Component):

    @property
    def tmux(self):
        return TmuxFacade(self.state, self.socket, self.options)

    @property
    def socket(self):
        return self.vim.vars.p('tmux_socket') | None

    @property
    def native_server(self):
        return NativeServer(socket_name=self.socket)

    @property
    def server(self):
        return Server(self.native_server)

    @property
    def new_state(self):
        return lambda: TmuxState(watcher=Just(self.create_watcher()))

    def create_watcher(self):
        interval = self.vim.vars.p('tmux_watcher_interval') | 1.0
        return Watcher(self, interval=interval)

    @property
    def watcher(self) -> Maybe[Watcher]:
        return self.state.watcher

    def record_lens(self, tpe, name):
        return (
            self.state.pane_lens_ident(name)
            if tpe == 'pane' else
            self.state.layout_lens_ident(name)
            if tpe == 'layout' else
            Empty()
        )

    @may_handle(Stage1)
    def stage_1(self):
        ''' Initialize the state. If it doesn't exist, Env will create it using *Tmux.new_state*
        '''
        default = self.vim.pflags.tmux_use_defaults.maybe(TmuxLoadDefaults())
        msgs = List(AddDispatcher(), self.with_sub(self.state), TmuxFindVim(), UpdateVimWindow())
        watcher = (Nil if self.vim.vars.p('tmux_no_watcher').true else List(StartWatcher()))
        return msgs + default.to_list + watcher

    @may_handle(Initialized)
    def initialized(self):
        return self.with_sub(self.state.set(initialized=True))

    @may_handle(AddDispatcher)
    def add_dispatcher(self):
        return self.data + TmuxDispatcher()

    @may_handle(TmuxFindVim)
    def find_vim(self):
        id = (
            self.vim.vars.pi('tmux_force_vim_pane_id')
            .o(lambda: self.native_vim_pane // _.id_i)
        )
        wid = (
            self.vim.vars.pi('tmux_force_vim_win_id')
            .o(lambda: self.native_vim_pane // _.window_id_i)
            .to_either('no vim win id')
        )
        sid = (
            self.vim.vars.pi('tmux_force_vim_session_id')
            .o(lambda: self.native_vim_pane // _.session_id_i)
            .to_either('no vim session id')
        )
        vim_w = self.vim.vars.p('tmux_vim_width').or_else(Just(88))
        pane = VimPane(id=id.to_maybe, window_id=wid, session_id=sid)
        vim_layout = VimLayout(name=VimPane.pane_name,
                               direction=LayoutDirections.vertical,
                               panes=List(pane), min_size=vim_w,
                               weight=Just(0.0001))
        root = Layout(name='root',
                      direction=LayoutDirections.horizontal,
                      layouts=List(vim_layout))
        win = VimWindow(id=wid.to_maybe, root=root)
        session = VimSession(id=sid.to_maybe, windows=List(win))
        new_state = self.state.append1.sessions(session)
        return self.with_sub(new_state)

    @may_handle(TmuxLoadDefaults)
    def load_defaults(self):
        main = TmuxCreateLayout('main',
                                options=Map(parent='root', weight=1))
        make = TmuxCreatePane('make', options=Map(parent='main', pin=True,
                                                  position=-1))
        return main, make

    @handle(StartWatcher)
    def start_watcher(self):
        return self.watcher / (lambda a: UnitIO(IO.delay(a.start)))

    @handle(TmuxCreateSession)
    def create_session(self):
        return (
            self.msg.options.get('name') /
            (lambda n: Session(id=n)) /
            List /
            self.state.append.sessions /
            self.with_sub
        )

    @may_handle(TmuxSpawnSession)
    def spawn_session(self):
        s = self.state.session(self.msg.id)
        self.server.new_session(session_name=s.x)

    @handle(TmuxCreateLayout)
    def create_layout(self):
        opts = self.msg.options
        dir_s = opts.get('direction') | 'vertical'
        direction = LayoutDirections.parse(dir_s)
        return (
            self._from_opt(Layout, name=self.msg.name, direction=direction) //
            L(self.tmux.add_layout_to_layout)(opts.get('parent'), _) /
            self.with_sub
        )

    @handle(TmuxCreatePane)
    def create_pane(self):
        opts = self.msg.options
        return (
            self._from_opt(Pane, name=self.msg.name) //
            L(self.tmux.add_pane_to_layout)(opts.get('parent'), _) /
            self.with_sub
        )

    @may_handle(TmuxOpen)
    def open(self):
        name = self.msg.name
        opt = self.msg.options
        pinned = opt.get('pinned').exists(I)
        def open():
            return (
                self.tmux.open_pane_ppm(name),
                (Nop() if pinned else TmuxPostOpen(name, opt - 'pinned').pub)
            )
        return (
            open()
            if self.tmux.view_exists(name) else
            Error(invalid_pane_name.format(name))
        )

    @may_handle(TmuxPostOpen)
    def open_pinned(self):
        return (
            self.tmux.pinned_panes(self.msg.ident) /
            L(TmuxOpen)(_, Map(pinned=True))
        ) + List(TmuxFixFocus(self.msg.ident), TmuxPack().pub.at(1))

    @handle(TmuxFixFocus)
    def fix_focus(self):
        return self.tmux.fix_focus(self.msg.pane, VimPane.pane_name) / UnitIO

    @handle(TmuxFocus)
    def focus(self):
        return self.tmux.focus(self.msg.pane) / UnitIO

    @handle(TmuxClosePane)
    def close(self):
        return (self.tmux.close_pane_ppm(self.msg.name) &
                Right(TmuxPack().pub.at(1)))

    @handle(TmuxRunCommand)
    def run_command(self) -> Maybe[Message]:
        job = self.msg.job
        options = self.msg.options
        shell = (options.get('shell')
                 .o(job.shell) //
                 self.data.shell)
        return shell.to_maybe.cata(
            L(self._run_in_shell)(job, _, options),
            L(self._run_command)(job, options)
        )

    @handle(TmuxRebootCommand)
    def tmux_reboot_command(self):
        return ((self.msg.job.command.target / TmuxClosePane) &
                Just(TmuxRunCommand(self.msg.job)))

    @handle(SetShellTarget)
    def set_shell_target(self) -> Maybe[Message]:
        target = self._pane(self.msg.target) / _.uuid
        setter = __.modify(__.set(target=target))
        msg = 'cannot set shell target {} for {}: pane not found'
        return (
            target
            .to_either(msg.format(self.msg.target, self.msg.shell.name)) //
            (lambda a: self.data.command_lens(self.msg.shell.uuid)) /
            setter
        )

    @handle(TmuxRunLineInShell)
    def run_line_in_shell(self):
        opt = self.msg.options
        return (
            opt.get('target') /
            (lambda a: Command(name='temp', line=a)) /
            L(self._run_in_shell)(_, self.msg.shell, opt)
        )

    @may_handle(TmuxRunTransient)
    def run_transient(self):
        return TmuxRunCommand(self.msg.cmd, options=self.msg.options)

    @may_handle(ViewPathMod)
    def view_path_mod(self):
        return self._run_vpm(self.msg)

    @may_handle(PanePathMod)
    def pane_path_mod(self):
        return self._run_vpm(self.msg)

    @may_handle(LayoutPathMod)
    def layout_path_mod(self):
        return self._run_vpm(self.msg)

    @may_handle(Quit)
    def quit(self):
        return QuitWatcher(), UnitIO(self.tmux.close_all)

    @handle(QuitWatcher)
    def quit_watcher(self):
        return self.watcher / (lambda a: UnitIO(IO.delay(a.stop)))

    @may_handle(TmuxInfo)
    def info(self):
        self.log.info('\n'.join(format_state(self.state)))

    @may_handle(Terminated)
    def terminated(self):
        return self.tmux.pane_mod(self.msg.pane.uuid, __.set(pid=Empty()))

    @may_handle(TmuxPack)
    def pack(self):
        return UnitIO(self.tmux.pack_sessions), UpdateVimWindow()

    @handle(UpdateVimWindow)
    def update_vim_window(self):
        def update_vim_window(sess, win_lens):
            win = win_lens.get()
            w = self.tmux.native_window(sess, win)
            return (
                (w / _.size) /
                Just /
                win.setter.size /
                win_lens.set /
                self.with_sub
            )
        return ((self.state.vim_session & self.state.vim_window_lens)
                .flat_map2(update_vim_window))

    @handle(Resized)
    def resized(self):
        def check(win, adapter):
            return win.size.contains(adapter.size).no.maybe(TmuxPack())
        return (
            self.state.initialized
            .flat_maybe((self.state.vim_window & self.tmux.native_vim_window))
            .flat_map2(check)
        )

    @may_handle(TmuxMinimize)
    def minimize(self):
        return self._minimize(__.set(minimized=True))

    @may_handle(TmuxRestore)
    def restore(self):
        return self._minimize(__.set(minimized=False))

    @may_handle(TmuxToggle)
    def toggle(self):
        return self._minimize(lambda a: a.set(minimized=not a.minimized))

    @handle(TmuxOpenOrToggle)
    def open_or_toggle(self):
        name = self.msg.pane
        return (
            self.tmux.view_open(name) /
            __.c(TmuxToggle(name), TmuxOpen(name))
        )

    @may_handle(TmuxKill)
    def kill(self):
        signals = self.msg.options.get('signals') | default_signals
        return self.tmux.kill_process(self.msg.pane, signals) / __.lmap(Fatal)

    @may_handle(TmuxPostCommand)
    def post_command(self):
        log_path = self._pane(self.msg.pane_ident) // _.log_path
        return CommandExecuted(self.msg.job, log_path).pub

    def _minimize(self, f):
        return self.tmux.view_mod(self.msg.pane, f), TmuxPack().pub.at(1)

    def _state_task(self,
                    f: Callable[[TmuxState], IO[Either[str, TmuxState]]],
                    options=Map()):
        cb = lambda d: f(self._state(d)) / __.map(L(self._with_sub)(d, _))
        return DataIO(_ // cb)

    def _run_vpm(self, vpm):
        return self._state_task(vpm.run, self.msg.options)

    def _pane(self, ident: Ident):
        return self.state.pane(ident)

    @property
    def _vim_pane(self):
        return self._pane(VimPane.pane_name)

    def _pane_uuid(self, ident: Ident):
        return self._pane(ident) / _.uuid

    def _default_pane_name(self):
        return 'make'

    @property
    def _vim_pid(self):
        return self.vim.call('getpid').to_either('no pid')

    @lazy
    def native_vim_pane(self):
        return self._vim_pid // self._find_vim_pane_by_pid

    def _find_vim_pane_by_pid(self, vim_pid: int) -> Maybe[PaneData]:
        return (
            self.server.pane_data
            .find(__.command_pids.exists(__.contains(vim_pid)))
        )

    def _run_command(self, job: CommandJob, opt):
        args = opt.get('args') | job.args
        return (
            job.line
            .resolve(self.vim, *args) /
            L(self._run_command_line)(job, _, opt)
        )

    def _run_command_line(self, job: CommandJob, line, opt):
        command = job.command
        kill = opt.get('kill') | command.kill
        signals = opt.get('signals') / List.wrap | command.signals
        pane_ident = (opt.get('target').o(command.target) |
                      self._default_pane_name)
        in_shell = Boolean('shell' in opt)
        def watch(path):
            msg = WatchCommand(job, path.view)
            return self.watcher / (lambda a: IO.delay(a.send, msg)) | IO.pure(None)
        runner, is_open = self.tmux.run_command_ppm(pane_ident, line, in_shell,
                                                    kill, signals)
        post = is_open.no.m(TmuxPostOpen(pane_ident, opt)).to_list
        return post.cons(runner % watch).cat(TmuxPostCommand(job, pane_ident))

    def _run_in_shell(self, job: CommandJob, shell: Shell, options: Map):
        ident = shell.target | self._default_pane_name
        opt = options + ('target', ident) + ('shell', shell)
        return (self._run_command(job, opt) /
                __.cons(TmuxRunCommand(CommandJob(command=shell), Map())))

    def _main_command(self, cmd_ident: Ident):
        holder = lambda cmd: cmd.shell // self.data.shell | cmd
        return self.data.command(cmd_ident) / holder


__all__ = ('Tmux',)



from typing import Union, Callable

from lenses import lens

from amino import Map, _, __

from ribosome.record import maybe_field, field
from ribosome.nvim import NvimFacade, AsyncVimProxy
from ribosome.machine.state import AutoData

from myo.command import Commands, Command
from myo.dispatch import Dispatchers, Dispatcher
from myo.util import Ident


class Env(AutoData):
    vim_facade = maybe_field((NvimFacade, AsyncVimProxy))
    initialized = field(bool, initial=False)
    commands = field(Commands, initial=Commands())
    dispatchers = field(Dispatchers, initial=Dispatchers())

    @property
    def vim(self) -> NvimFacade:
        return self.vim_facade | (lambda: NvimFacade(None, 'no vim was set in `Env`'))

    def cat(self, item: Union[Command, Dispatcher]):
        name = (
            'commands' if isinstance(item, Command) else
            'dispatchers' if isinstance(item, Dispatcher) else
            None
        )
        return self.mod(name, _ + item)

    __add__ = cat

    def command(self, name: str):
        return self.commands[name]

    def command_or_job(self, name: str):
        return self.commands.command_or_job(name)

    def shell(self, name: str):
        return self.commands.shell(name)

    def dispatch_message(self, cmd: Command, options: Map):
        return self.dispatchers.message(cmd, options)

    def _lens(self, getter: Callable, ident: Ident):
        return (
            getter(getter(self))
            .find_lens_pred(__.has_ident(ident))
            .to_either('no command with ident {} found'.format(ident)) /
            getter(getter(lens(self))).add_lens
        )

    def command_lens(self, ident: Ident):
        return self._lens(_.commands, ident)

    def transient_command(self, cmd: Command):
        return self.mod('commands', __.transient_command(cmd))

__all__ = ('Env',)



from ribosome.util.callback import VimCallback

from amino import _


class DuplicatesFilter(VimCallback):

    def __call__(self, result):
        return result.modder.events(_.distinct)

__all__ = ('DuplicatesFilter',)



import re
from uuid import UUID

from amino import List, Maybe, Empty, L, _

from ribosome.record import field

from myo.util import parse_id

_session_id_re = re.compile('^\$(\d+)$')


def parse_session_id(value):
    return parse_id(value, _session_id_re, 'session')

_win_id_re = re.compile('^@(\d+)$')


def parse_window_id(value):
    return parse_id(value, _win_id_re, 'window')

_pane_id_re = re.compile('^%(\d+)$')


def parse_pane_id(value):
    return parse_id(value, _pane_id_re, 'pane')


def format_pane(pane):
    return List(pane.desc)


def format_layout(lay):
    return (List(lay.desc) + indent(lay.layouts // format_layout) +
            indent(lay.panes // format_pane))


def indent(lines):
    return lines / '  {}'.format


def format_window(win):
    return List(win.desc) + indent(format_layout(win.root))


def format_state(state) -> List[str]:
    return state.all_windows // format_window

__all__ = ('parse_window_id', 'parse_pane_id', 'format_state', 'format_window',
           'indent', 'format_layout', 'format_pane', 'parse_session_id')
